"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var fs_extra_1 = require("fs-extra");
var walk = require("klaw");
var Promise = require("bluebird");
var path = require("path");
var api_1 = require("./api");
var utils_1 = require("./utils");
var memory_fs_1 = require("./memory-fs");
var ensureDir = Promise.promisify(fs_extra_1.ensureDir);
var readFile = Promise.promisify(fs_extra_1.readFile);
var writeFile = Promise.promisify(fs_extra_1.writeFile);
var remove = Promise.promisify(fs_extra_1.remove);
var rmdir = Promise.promisify(fs_extra_1.rmdir);
var stat = Promise.promisify(fs_extra_1.stat);
var access = Promise.promisify(fs_extra_1.access);
// TODO extract chokidar watch mechanism to configuration
var LocalFileSystemCrudOnly = (function () {
    function LocalFileSystemCrudOnly(baseUrl, ignore) {
        this.baseUrl = baseUrl;
        this.events = utils_1.makeEventsEmitter();
        this.ignore = [];
        this.isIgnored = function (path) { return false; };
        if (ignore) {
            this.isIgnored = utils_1.getIsIgnored(ignore);
        }
        ;
    }
    LocalFileSystemCrudOnly.prototype.saveFile = function (relPath, newContent) {
        if (this.isIgnored(relPath)) {
            return Promise.reject(new Error("Unable to save ignored path: '" + relPath + "'"));
        }
        var pathArr = utils_1.getPathNodes(relPath);
        var name = pathArr.pop() || '';
        var fullPath = path.join.apply(path, [this.baseUrl].concat(pathArr));
        return ensureDir(fullPath)
            .then(function () { return writeFile(path.join(fullPath, name), newContent); });
    };
    LocalFileSystemCrudOnly.prototype.deleteFile = function (relPath) {
        if (!relPath) {
            return Promise.reject(new Error("Can't delete root directory"));
        }
        if (this.isIgnored(relPath)) {
            return Promise.resolve();
        }
        var fullPath = path.join.apply(path, [this.baseUrl].concat(utils_1.getPathNodes(relPath)));
        return access(fullPath)
            .then(function () { return stat(fullPath); }, function (err) { return null; })
            .then(function (stats) {
            if (stats) {
                if (stats.isFile()) {
                    return remove(fullPath);
                }
                else {
                    throw new Error("not a file: " + relPath);
                }
            }
        });
    };
    LocalFileSystemCrudOnly.prototype.deleteDirectory = function (relPath, recursive) {
        var pathArr = utils_1.getPathNodes(relPath);
        if (pathArr.length === 0) {
            return Promise.reject(new Error("Can't delete root directory"));
        }
        if (this.isIgnored(relPath)) {
            return Promise.resolve();
        }
        var fullPath = path.join.apply(path, [this.baseUrl].concat(pathArr));
        return access(fullPath)
            .then(function () { return stat(fullPath); }, function (err) { return null; })
            .then(function (stats) {
            if (stats) {
                if (stats.isDirectory()) {
                    return recursive ? remove(fullPath) : rmdir(fullPath);
                }
                else {
                    throw new Error("not a directory: " + relPath);
                }
            }
        });
    };
    LocalFileSystemCrudOnly.prototype.loadTextFile = function (relPath) {
        if (this.isIgnored(relPath)) {
            return Promise.reject(new Error("Unable to read ignored path: '" + relPath + "'"));
        }
        return readFile(path.join(this.baseUrl, relPath), 'utf8');
    };
    LocalFileSystemCrudOnly.prototype.loadDirectoryTree = function () {
        var _this = this;
        // using an in-memory instance to build the result
        var promises = [];
        var memFs = new memory_fs_1.MemoryFileSystem();
        return Promise.fromCallback(function (callback) {
            var _a = _this, baseUrl = _a.baseUrl, isIgnored = _a.isIgnored;
            walk(baseUrl)
                .on('readable', function () {
                var item;
                while ((item = this.read())) {
                    var itemPath = path.relative(baseUrl, item.path).split(path.sep).join(api_1.pathSeparator);
                    if (isIgnored(itemPath)) {
                        return;
                    }
                    else if (item.stats.isDirectory()) {
                        promises.push(memFs.ensureDirectory(itemPath));
                    }
                    else if (item.stats.isFile()) {
                        promises.push(memFs.saveFile(itemPath, ''));
                    }
                    else {
                        console.warn("unknown node type at " + itemPath, item);
                    }
                }
            })
                .on('end', function () {
                Promise.all(promises)
                    .then(function () { return memFs.loadDirectoryTree(); })
                    .then(callback.bind(null, null));
            });
        });
    };
    LocalFileSystemCrudOnly.prototype.ensureDirectory = function (relPath) {
        if (this.isIgnored(relPath)) {
            return Promise.reject(new Error("Unable to read and write ignored path: '" + relPath + "'"));
        }
        var pathArr = utils_1.getPathNodes(relPath);
        var fullPath = path.join.apply(path, [this.baseUrl].concat(pathArr));
        return ensureDir(fullPath);
    };
    return LocalFileSystemCrudOnly;
}());
exports.LocalFileSystemCrudOnly = LocalFileSystemCrudOnly;
//# sourceMappingURL=local-fs-crud-only.js.map